#ifndef SCD_MAP_CPP
#define SCD_MAP_CPP


#include <map.h>
#include <stdio.h>

class SCD_Map
{

private:
	map<char*, void*> m;
	map<char*, void*>::iterator it;

public:
	void* get(char *key, bool *p_success = NULL)
	{
		map<char*, void*>::iterator p;

/*
		p = m.begin();
		while( p != m.end() )
		{
			if( strcmp( key , p->first ) == 0 )	break;
			p++;
		}
*/
		p = m.find(key);

		void *ret = NULL;
		if( p != m.end() )	ret = p->second;
		if(p_success)	*p_success = ( p != m.end() );

		return ret;
	}
	void put(char *key, void *value)
	{
		m.insert( pair<char*, void*>(key, value) );
	}
	bool containsKey(char *key)
	{
		map<char*, void*>::iterator p;
		p = m.find(key);
		return ( p != m.end() );
	}
	bool rename(char *oldname, char *newname)
	{
		void *value;
		if( ! containsKey(newname) ) return false;
		if( ! remove( oldname , value ) )	return false;
		put( newname , value );
		return true;
	}
	bool remove(char *key, void *value = NULL)
	{
		map<char*, void*>::iterator p;
		p = m.find(key);
		if( p != m.end() )
		{
			m.erase(p);
			if( ! value )	value = p->second;
			return true;
		}
		else return false;
	}
	int  size()		{	return m.size();	}
	bool isEmpty()	{	return m.empty();	}
	void clear()	{	m.clear();	}

	void init_iterator()	{	it = m.begin();	}
	bool hasNext()			{	return it != m.end();	}
	void* next()	{	void *ret;	ret = it->first;	it++;	return ret;	}
};





#include <stdio.h>

typedef struct SCD_Config_folder
{
	SCD_Map folders , keys;
	SCD_Config_folder *up;

} SCD_Config_folder;

class SCD_Config
{

private:
	SCD_Config_folder  root , *current;


	SCD_Config_folder* get_folder(char *path)
	{
		SCD_Config_folder *result = current;

		if( path == NULL )	return NULL;
		if( *path == '/' )	{	result = & root;	path ++;	}

		bool flag = true;

		while( *path != '\0' && flag )
		{
			char *tmp;
			tmp = strchr( path , '/' );
			if( tmp != NULL )	*tmp = '\0';

			SCD_Config_folder *tf;
			tf = (SCD_Config_folder*) result->folders.get(path);
			if( tf == NULL )	flag = false;
			else result = tf;

			if( tmp != NULL )	*tmp = '/';
			else break;

			path = tmp + 1;
		}

		if(flag)	return result;
		else return NULL;
	}

public:
	SCD_Config()	{	root.up = NULL;	current = & root;	}
	~SCD_Config()	{	printf("SCD_Config destructor\r\n");	clear();	}

	bool load(FILE *fp , SCD_Config_folder *folder , char *buf )
	{

		char *name, *value;
		SCD_Config_folder *nf;
		while(true)
		{
			if( ! fgets( buf , 1000 , fp ) )	return false;
			switch( buf[0] )
			{
				case 'k' :	//define key
					if( buf[1] == '\0' )	return false;
					name = buf + 2;

					value = strchr( name , '=' );
					if( ! value )	return false;
					*value = '\0';	value ++;

					folder->keys.put( name , value );

					break;
				case 'f' :	//define folder
					if( buf[1] == '\0' )	return false;
					name = buf + 2;
					if( *name == '\0' )	return false;

					nf = new SCD_Config_folder;
					folder->folders.put( name , nf );

					load( fp , nf , buf );

					break;
				case 'c' :	//close folder
					if( buf[1] == '\0' )	return false;
					return true;
				default :	//ignore
					break;
			}
		}

		return true;
	}

	bool load(char *filename)
	{
		FILE *fp;
		fp = fopen(filename, "rb");
		if( fp == NULL )	return false;

		cd("/");
		clear();

		char buf[1001];
		load( fp , & root , buf );

		fclose(fp);
		return true;
	}
	bool save(char *filename)
	{
		FILE *fp;
		if( filename == NULL )	fp = stdout;
		else fp = fopen(filename, "wb");
		if( fp == NULL )	return false;

		save(fp, & root );

		fclose(fp);
		return true;
	}
	void save(FILE *fp , SCD_Config_folder *folder )
	{
		SCD_Map *km , *fm;
		km = & (folder->keys);
		fm = & (folder->folders);

		km->init_iterator();
		while( km->hasNext() )
		{
			char *name = (char*) km->next();
			char *value = (char*) km->get(name);
			if( ! value )	value = "";
			fprintf(fp, "k %s=%s\r\n", name , value );
		}

		fm->init_iterator();
		while( fm->hasNext() )
		{
			char *name = (char*) fm->next();
			fprintf( fp, "f %s\r\n", name );
			save(fp, (SCD_Config_folder*) fm->get(name) );
			fprintf( fp , "c\r\n" );
		}
	}

	bool cd(char *path)
	{
		SCD_Config_folder* f;
		f = get_folder(path);
		if( f == NULL )	return false;
		current = f;
		return true;
	}
	bool cdup()
	{
		if( current->up == NULL )	return false;
		current = current->up;
		return true;
	}
	void mkdir(char *name)
	{
		if( dir_exist(name) )	return;
		SCD_Config_folder *nf = new SCD_Config_folder;
		nf->up = current;
		current->folders.put( name , nf );
	}
	bool rename_dir(char *oldname , char *newname)
	{	return current->folders.rename(oldname, newname);	}
	bool dir_exist(char *path)
	{
		SCD_Config_folder* f;
		f = get_folder(path);
		return ( f != NULL );
	}
	bool rmdir(char *name)
	{
		SCD_Config_folder *tmp_current, *up_folder, *t;

		tmp_current = current;

		t = get_folder(name);
		if( t == NULL ) return false;
		current = t;

		t->folders.init_iterator();
		while( t->folders.hasNext() )
			rmdir( (char*) t->folders.next() );

		clearKeys();

		current = tmp_current;

		up_folder = current->up;
		if( up_folder )	up_folder->folders.remove(name);

printf("clear folder %s\r\n", name);
		return true;
	}
	void clearDirs()
	{
		current->folders.init_iterator();
		while( current->folders.hasNext() )
			rmdir( (char*) current->folders.next() );
		current->folders.clear();
	}


	char* get(char *key, bool *is_success = NULL )
	{	return (char*) current->keys.get(key, is_success);	}
	void put(char *key, char *value)
	{	current->keys.put(key, value);	}
	bool containsKey(char *key)
	{	current->keys.containsKey(key);   }
	bool renameKey(char *oldname, char *newname)
	{	return current->keys.rename(oldname, newname);	}
	bool deleteKey(char *key)
	{	return current->keys.remove(key) != NULL;	}
	void clearKeys()
	{
		current->keys.clear();
printf("clear keys\r\n");
	}

	void clear()
	{
		cd("/");
		clearKeys();
		clearDirs();
	}
};



int main()
{
/*
	SCD_Map m;
	m.put( (char*)"1", (char*)"one");
	m.put( (char*)"3", (char*) NULL);
	m.put( (char*)"9", (char*)"nine");
	m.put( (char*)"5", (char*)"five");
	m.init_iterator();
	while( m.hasNext() )
		printf( "%s\n", (char*)m.next() );
*/

	SCD_Config *cfg = new SCD_Config();

	cfg->put("1", "one");
	cfg->put("7", NULL);
	cfg->put("3", "three");
	cfg->put("9", "nine");

	cfg->mkdir("favorite");
	cfg->cd("favorite");
	cfg->put("z", "desert.twbbs.org:23");
	cfg->put("ߪ", "140.119.164.150:23");
	cfg->put("eden", "eden.cs.nccu.edu.tw:23");

	cfg->mkdir("homepage");
	cfg->cd("homepage");
	cfg->put("nccu web" , "http://www.nccu.edu.tw/");
	cfg->put("nccu cs", "http://www.cs.nccu.edu.tw");

	cfg->cdup();
	cfg->mkdir("test");

	cfg->cd("/");
	cfg->mkdir("option");
	cfg->cd("option");
	cfg->put("timeout", "100");
	cfg->put("defaultpath", "/");

	cfg->cd("/");
	cfg->deleteKey("1");
	cfg->rmdir("favorite");

	cfg->save(NULL);
	cfg->save("log.txt");

	delete cfg;
/*
	m.clear();
	printf( m.containsKey("one") ? "true" : "false" );
	printf( m.containsKey("three") ? "true" : "false" );
	printf("\n%d\n", m.size() );
//	printf( (char*)m.get("one") );
*/
}

#endif
